
/*#define TUEDACS*/
/*#define PIZZADEBUG*/
#define PIZZAMODEL
/*#define PIZZAVIEW*/
/*#define IS_WIN32*/
#define FASTHOME

/*#define MASTER*/
/*#define SLAVE*/

/*
	PIZZA10 (Matlab R11/R12 version)

	(c) Rene' van de Molengraft, 2001-2004

	last update: April, 27th, 2004

	Inputs : u[0]...u[3]   = motor commands

		 u[4]          = start signal
			           1 = start

		 u[5]	       = not used

	States : x[0]...x[3]   = velocity filter states

		 x[4]...x[7]   = integrators for I-action

		 x[8]...x[15]  = states of pizza model (only #ifdef PIZZAMODEL)

	Outputs: y[0]...y[3]   = motor positions

		 y[4]          = status flag takes one of the following values:
					WAITING_FOR_START
					HOMING
					MOVING
					READY
					P1_POS_VIOLATION
					P2_POS_VIOLATION
					P3_POS_VIOLATION
					P4_POS_VIOLATION
					P1_VEL_VIOLATION
					P2_VEL_VIOLATION
					P3_VEL_VIOLATION
					P4_VEL_VIOLATION
					P1_CLIPPING
					P2_CLIPPING
					P3_CLIPPING
					P4_CLIPPING
 
		 y[5]          = time from initialization

		 y[6]...y[9]   = not used

		 y[10]...y[17] = states of pizza model (only #ifdef PIZZAMODEL)
*/

#define S_FUNCTION_NAME pizza10m
#define S_FUNCTION_LEVEL 2

#include "simstruc.h"

#define U(element) (*uPtrs[element])  /* Pointer to Input Port0 */

#define NINPUTS		6
#ifdef PIZZAMODEL
#define NSTATES		16
#define NOUTPUTS	18
#else
#define NSTATES		8
#define NOUTPUTS	10
#endif
#define NRWRK		400
#define NIWRK		100

/* encoder scaling factors */
#define FACT1		8.7989e-006
#define FACT2		7.5209e-006
#define FACT3		1.6886e-005
#define FACT4		5.0000e-006

#define HOME1		-0.0232
#define HOME2		0.0269
#define HOME3		0.2892
#define HOME4		0.0606

#define PI		3.141592653

/* aliases for global variables in work-space */

#define pizza_status		prwrk[piwrk[0]]
#define pizza_pos(element)	prwrk[piwrk[1]+element]
#define pizza_fact(element)	prwrk[piwrk[2]+element]
#define pizza_vel(element)	prwrk[piwrk[3]+element]
#define us(element)		prwrk[piwrk[4]+element]
#define pizza_result		prwrk[piwrk[5]]
#define pizza_time		prwrk[piwrk[6]]
#define target(element)		prwrk[piwrk[7]+element]
#define maxpos(element)		prwrk[piwrk[8]+element]
#define minpos(element)		prwrk[piwrk[9]+element]
#define margin(element)		prwrk[piwrk[10]+element]
#define p_airbag(element)	prwrk[piwrk[11]+element]
#define d_airbag(element)	prwrk[piwrk[12]+element]
#define bag(element)		prwrk[piwrk[13]+element]
#define maxout(element)		prwrk[piwrk[14]+element]
#define clip(element)		prwrk[piwrk[15]+element]
#define kp(element)		prwrk[piwrk[16]+element]
#define kd(element)		prwrk[piwrk[17]+element]
#define ki(element)		prwrk[piwrk[18]+element]
#define vh(element)		prwrk[piwrk[19]+element]
#define epd(element)		prwrk[piwrk[20]+element]
#define tset(element)		prwrk[piwrk[21]+element]
#define xset(element)		prwrk[piwrk[22]+element]
#define aset(element)		prwrk[piwrk[23]+element]
#define bset(element)		prwrk[piwrk[24]+element]
#define cset(element)		prwrk[piwrk[25]+element]
#define homdir(element)		prwrk[piwrk[26]+element]
#define epdtar(element)		prwrk[piwrk[27]+element]
#define movdir(element)		prwrk[piwrk[28]+element]
#define iaxis(element)		prwrk[piwrk[29]+element]
#define fric(element)		prwrk[piwrk[30]+element]
#define visc(element)		prwrk[piwrk[31]+element]
#define gain(element)		prwrk[piwrk[32]+element]
#define pizza_rest		prwrk[piwrk[33]]
#define JOG_IDX			34

#include "math.h"

#ifdef TUEDACS
int_T td_dac_write_chan(real_T, int_T, int_T, int_T);
int_T td_enc_clear_chan(int_T, int_T, int_T); 
int_T td_enc_read_chan(real_T*, int_T, int_T, int_T);
int_T td_enc_ref(int_T*, int_T, int_T);
int_T td_do_select_bit(int_T,int_T,int_T);
int_T td_do_write_bit(real_T,int_T,int_T,int_T);
int_T td_di_read_bit(real_T*,int_T,int_T,int_T);
#define TD_DIRECT 0
#define TD_CACHED 1
#define TD_STORED 2
#endif

#include "pizza.h"
#include "pizza_work.c"
#include "pizza_jog.c"
#include "pizza_pid.c"
#include "pizza_papi.c"
#include "pizza_airbag.c"
#include "pizza_motor.c"





#ifdef IS_WIN32
#include <windows.h>
void StartupProgram(char* pappname)
{
	STARTUPINFO startupinfo;
	PROCESS_INFORMATION pinfo;

	startupinfo.cb=sizeof(STARTUPINFO);
	startupinfo.lpReserved=NULL;
	startupinfo.lpDesktop=NULL;
	startupinfo.lpTitle=NULL;
	startupinfo.dwFlags=STARTF_USESHOWWINDOW;
	startupinfo.cbReserved2=0;
	startupinfo.lpReserved2=NULL;
	startupinfo.wShowWindow=SW_SHOW;
	CreateProcess(NULL,pappname,
        	      NULL,NULL,0,0,
		      NULL,NULL,&startupinfo,&pinfo);
}
#endif





/*====================*
 * S-function methods *
 *====================*/

static void mdlInitializeSizes(SimStruct *S)
{
        int i;

        ssSetNumContStates(S,NSTATES);
	ssSetNumDiscStates(S,0);

	if (!ssSetNumInputPorts(S,1)) return;
	ssSetInputPortWidth(S,0,NINPUTS);
#ifdef  MATLAB_MEX_FILE
	ssSetInputPortDirectFeedThrough(S,0,NINPUTS);
#else
	ssSetInputPortDirectFeedThrough(S,0,0);
#endif

	if (!ssSetNumOutputPorts(S,1)) return;
	ssSetOutputPortWidth(S,0,NOUTPUTS);

	ssSetNumSampleTimes(S,1);
	ssSetNumRWork(S,NRWRK);
	ssSetNumIWork(S,NIWRK);
	ssSetNumPWork(S,0);
	ssSetNumModes(S,0);
	ssSetNumNonsampledZCs(S,0);

#ifdef TUEDACS
/*	initialize hardware */

	for (i=0;i<4;i++) {
		td_do_select_bit(4+i,0,TD_DIRECT);
	}

/*	pulse for power on */
	td_do_write_bit(0.0,6,0,TD_DIRECT);
	usleep(500000);
	td_do_write_bit(1.0,6,0,TD_DIRECT);
	usleep(500000);
	td_do_write_bit(0.0,6,0,TD_DIRECT);

/*	enable servo amplifiers */
	td_do_write_bit(0.0,4,0,TD_DIRECT);
	td_do_write_bit(1.0,5,0,TD_DIRECT);
#endif

}





static void mdlInitializeSampleTimes(SimStruct *S)
{
    ssSetSampleTime(S,0,CONTINUOUS_SAMPLE_TIME);
    ssSetOffsetTime(S,0,0.0);
}





#define MDL_INITIALIZE_CONDITIONS
static void mdlInitializeConditions(SimStruct *S)
{
	real_T *x=ssGetContStates(S);

	int_T *piwrk=ssGetIWork(S);
	real_T *prwrk=ssGetRWork(S);

	int_T i;
	real_T step_size;

/*	check step size */

	step_size=ssGetStepSize(S);
	if (step_size!=0.002) {
	  ssSetErrorStatus(S, "Step size must be 0.002 s.");
	}

/*	initialize global variables in work-space */

	rwrk_init_all(S);

/*	initial values of global variables */

/*	axis power on (1) or off (0) */

	iaxis(0)=1;
	iaxis(1)=1;
	iaxis(2)=1;
	iaxis(3)=1;

	movdir(0)=-1.0;
	movdir(1)=1.0;
	movdir(2)=1.0;
	movdir(3)=1.0;

	pizza_status=(real_T) WAITING_FOR_START;

	pizza_fact(0)=FACT1;
	pizza_fact(1)=FACT2;
	pizza_fact(2)=FACT3;
	pizza_fact(3)=FACT4;

	for (i=0;i<4;i++) {
	  us(i)=0.0;
	}

	pizza_time=0.0;

	target(0)=-0.1;
	target(1)=0.1;
	target(2)=2.4;
	target(3)=0.2;

	pizza_rest=1.0;

/*	airbag boundaries */

	maxpos(0)=-HOME1;
	maxpos(1)=0.607-HOME2;
	maxpos(2)=6.16-HOME3;
	maxpos(3)=0.64-HOME4;

	minpos(0)=-0.315-HOME1;
	minpos(1)=-HOME2;
	minpos(2)=-HOME3;
	minpos(3)=-HOME4;

	margin(0)=0.05;
	margin(1)=0.05;
	margin(2)=0.3;
	margin(3)=0.05;

/*	airbag parameters */

	p_airbag(0)=500.0;
	p_airbag(1)=400.0;
	p_airbag(2)=500.0;
	p_airbag(3)=500.0;

	d_airbag(0)=10;
	d_airbag(1)=10;
	d_airbag(2)=10;
	d_airbag(3)=10;

	bag(0)=(real_T) P1_POS_VIOLATION;
	bag(1)=(real_T) P2_POS_VIOLATION;
	bag(2)=(real_T) P3_POS_VIOLATION;
	bag(3)=(real_T) P4_POS_VIOLATION;

/*	clipping parameters */

	maxout(0)=2.5;
	maxout(1)=2.5;
	maxout(2)=2.5;
	maxout(3)=2.5;

	clip(0)=P1_CLIP;
	clip(1)=P2_CLIP;
	clip(2)=P3_CLIP;
	clip(3)=P4_CLIP;

/*	control parameters */

	kp(0)=800.0;
	kp(1)=400.0;
	kp(2)=500.0;
	kp(3)=800.0;

	kd(0)=8;
	kd(1)=4;
	kd(2)=8;
	kd(3)=20;

	ki(0)=5.0;
	ki(1)=1.0;
	ki(2)=0.1;
	ki(3)=2;

/*	homing parameters */

#ifndef FASTHOME
	vh(0)=0.04;
	vh(1)=0.04;
	vh(2)=0.2;
	vh(3)=0.04;
#else
	vh(0)=0.1;
	vh(1)=0.1;
	vh(2)=0.5;
	vh(3)=0.1;
#endif

#ifdef PIZZAMODEL

/*	pizza model parameters */

	fric(0)=0.3/0.48;
/*	fric(1)=0.95/0.24; */
	fric(1)=0.3/0.48;
	fric(2)=0.9/0.11;
	fric(3)=0.4/1.12;

	visc(0)=0.01;
	visc(1)=0.01;
	visc(2)=0.01;
	visc(3)=0.01;

	gain(0)=1/0.48;
	gain(1)=1/0.24;
	gain(2)=1/0.11;
	gain(3)=1/1.12;


#endif

/*	initialize pizza */

        pizza_initialize(&pizza_pos(0),S);

#ifdef PIZZAMODEL
/*	pizza model initial condition */
	x[8]=-0.1;
	x[9]=0.1;
	x[10]=1.0;
	x[11]=0.2;

	for (i=0;i<4;i++) {
	  pizza_pos(i)=x[8+i];
	}

/*	initial velocity */
	for (i=0;i<4;i++) {
	  x[12+i]=0.0;
	}
#endif

/*	controller initial condition */

	pid_ini(x,&pizza_pos(0));
}





static void mdlOutputs(SimStruct *S, int_T tid)
{
	real_T *y=ssGetOutputPortRealSignal(S,0);
	real_T *x=ssGetContStates(S);
	InputRealPtrsType uPtrs=ssGetInputPortRealSignalPtrs(S,0);

	int_T *piwrk=ssGetIWork(S);
	real_T *prwrk=ssGetRWork(S);

	int_T i,istat;
	real_T time,ttt;

/*	current time */

	time=ssGetT(S);

/*	current position */

#ifdef TUEDACS
	td_enc_read_chan(&ttt,0,0,TD_STORED);
	pizza_pos(0)=pizza_fact(0)*ttt;
	td_enc_read_chan(&ttt,1,0,TD_STORED);
	pizza_pos(1)=pizza_fact(1)*ttt;
	td_enc_read_chan(&ttt,0,1,TD_STORED);
	pizza_pos(2)=pizza_fact(2)*ttt;
	td_enc_read_chan(&ttt,1,1,TD_STORED);
	pizza_pos(3)=pizza_fact(3)*ttt;
#else
	for (i=0;i<4;i++) {
	  pizza_pos(i)=x[i+8];
	}
#endif

	istat=(int_T) pizza_status;
	switch (istat)
	{
	case WAITING_FOR_START:

/*	  switch off motor commands */

	  for (i=0;i<4;i++) {
	    us(i)=0.0;
	  }

/*	  gravity compensation */

	  us(0)=-1.0*pizza_rest;
	  pizza_rest=0.9995*pizza_rest;

/*	  zeroise outputs, except for status out */

	  for (i=0;i<NOUTPUTS;i++) {
	    y[i]=0.0;
	  }
	  for (i=0;i<4;i++) {
	    y[i]=pizza_pos(i);
	  }
	  y[4]=(real_T) WAITING_FOR_START;

	  pizza_result=(real_T) send_motor_command(&us(0),S);

	  for (i=0;i<4;i++) {
	    y[6+i]=us(i);
	  }

	  break;

	case HOMING:

	  pizza_homing_out(&us(0),&pizza_vel(0),x,&pizza_pos(0),time,S);

	  for (i=0;i<NOUTPUTS;i++) {
	    y[i]=0.0;
	  }
	  for (i=0;i<4;i++) {
	    y[i]=pizza_pos(i);
	  }
	  y[4]=(real_T) HOMING;

	  pizza_result=(real_T) send_motor_command(&us(0),S);

	  for (i=0;i<4;i++) {
	    y[6+i]=us(i);
	  }

	  break;

        case MOVING:

	  pizza_moving_out(&us(0),&pizza_vel(0),x,&pizza_pos(0),time,S);

	  for (i=0;i<NOUTPUTS;i++) {
	    y[i]=0.0;
	  }
	  for (i=0;i<4;i++) {
	    y[i]=pizza_pos(i);
	  }
	  y[4]=(real_T) MOVING;

	  pizza_result=(real_T) send_motor_command(&us(0),S);

	  for (i=0;i<4;i++) {
	    y[6+i]=us(i);
	  }

	  break;

	case READY:

/*	  pass user inputs */

	  for (i=0;i<4;i++) {
	    us(i)=U(i);
	  }

#ifdef PIZZADEBUG
	  mexPrintf("t = %f, us = [ %f   %f   %f   %f ]\n",time,us(0),us(1),us(2),us(3));
#endif

/*	  check for boundary tresspassing */

	  pizza_result=(real_T) airbag(&us(0),&pizza_pos(0),&pizza_vel(0),0,S);
	  if ((pizza_result==(real_T) P1_POS_VIOLATION)|
	      (pizza_result==(real_T) P2_POS_VIOLATION)|
	      (pizza_result==(real_T) P3_POS_VIOLATION)|
	      (pizza_result==(real_T) P4_POS_VIOLATION)) {
	    pizza_status=pizza_result;
	  }

/*	  outputs */

	  for (i=0;i<4;i++) {
	    y[i]=pizza_pos(i);
	  }
	  y[4]=(real_T) pizza_status;
	  y[5]=time-pizza_time;

	  pizza_result=(real_T) send_motor_command(&us(0),S);

	  for (i=0;i<4;i++) {
	    y[6+i]=us(i);
	  }

	  break;

	case P1_POS_VIOLATION:
	case P2_POS_VIOLATION:
	case P3_POS_VIOLATION:
	case P4_POS_VIOLATION:

	  pizza_result=(real_T) airbag(&us(0),&pizza_pos(0),&pizza_vel(0),1,S);

	  pizza_result=(real_T) send_motor_command(&us(0),S);

	  for (i=0;i<4;i++) {
	    y[6+i]=us(i);
	  }

	  break;

/*	case P1_CLIPPING:*/
/*	case P2_CLIPPING:*/
/*	case P3_CLIPPING:*/
/*	case P4_CLIPPING:*/

/*	  break; */

	case STOP:

	  pizza_result=(real_T) send_motor_command(&us(0),S);

	  for (i=0;i<4;i++) {
	    y[6+i]=us(i);
	  }

	  break;

	}

#ifdef PIZZAMODEL

	for (i=0;i<4;i++) {
	  y[6+i]=us(i);
	}

	for (i=0;i<8;i++) {
	  y[10+i]=x[8+i];
	}

#endif
}





#define MDL_UPDATE
static void mdlUpdate(SimStruct *S, int_T tid)
{
	real_T *x=ssGetContStates(S);
	InputRealPtrsType uPtrs=ssGetInputPortRealSignalPtrs(S,0);

	int_T *piwrk=ssGetIWork(S);
	real_T *prwrk=ssGetRWork(S);

	int_T istat,ival;
	real_T time;

/*	check for stop... */

#ifdef SLAVE
/*	check enable bit (read dio bit 2 of QAD link 0) */

	dio_update(0);
	dio_read(&ival,2,0);
	if (ival==0) {
#else
	if (U(4)==0) {
#endif

/*	  reinitialize pizza */

	  pizza_initialize(&pizza_pos(0),S);
	  pid_ini(x,&pizza_pos(0));

	  pizza_status=WAITING_FOR_START;

#ifdef MASTER
/*	  disable slave (reset dio bit 6 of QAD link 1) */

	  dio_write(0,6,1);
	  dio_update(1);

/*	  master not ready for sync (reset dio bit 7 of QAD link 1) */

	  dio_write(0,7,1);
	  dio_update(1);
#endif

#ifdef SLAVE
/*	  slave not ready for sync  (reset dio bit 7 of QAD link 1) */

	  dio_write(0,7,1);
	  dio_update(1);
#endif

	}

/*	current time */

	time=ssGetT(S);

	istat=(int_T) pizza_status;
	switch (istat)
	{
	case WAITING_FOR_START:

#ifdef SLAVE
/*	  check enable bit (read dio bit 2 of QAD link 0) */

	  dio_update(0);
	  dio_read(&ival,2,0);
	  if (ival==1) {
#else
	  if (U(4)==1) {
#endif
/*	    go home... */
	    pizza_status=HOMING;
	    pizza_start_homing(&pizza_pos(0),time,S);
	    pizza_rest=1.0;

#ifdef MASTER
/*	    enable slave (set dio bit 6 of QAD link 1 to one) */

	    dio_write(1,6,1);
	    dio_update(1);
#endif
	  }

	  break;

	case HOMING:

	  if (pizza_synchronize(&pizza_pos(0),&target(0),time,PIZZAHOME,S)) {
	    pizza_status=MOVING;
	    pizza_start_moving(&pizza_pos(0),time,S);
	  }

	  break;

        case MOVING:

	  if (pizza_synchronize(&pizza_pos(0),&target(0),time,PIZZAMOVE,S)) {
	    pizza_time=time;

#ifdef MASTER
	    pizza_status=WAITING_FOR_SLAVE;
#else
	    pizza_status=READY;
#endif

#ifdef SLAVE
	    pizza_status=WAITING_FOR_MASTER;
#endif
	  }

	  break;

#ifdef SLAVE
	case WAITING_FOR_MASTER:

/*	  slave ready for sync (set dio bit 7 of QAD link 1) */

	  dio_write(1,7,1);
	  dio_update(1);

/*	  check master sync bit (read dio bit 3 of QAD link 0) */
	  dio_update(0);
	  dio_read(&ival,3,0);
	  if (ival==1) {
	    pizza_time=time;
	    pizza_status=READY;
	  }

	break;
#endif

#ifdef MASTER
	case WAITING_FOR_SLAVE:

/*	  master ready for sync (set dio bit 7 of QAD link 1) */

	  dio_write(1,7,1);
	  dio_update(1);

/*	  check slave sync bit (read dio bit 3 of QAD link 0) */
	  dio_update(0);
	  dio_read(&ival,3,0);
	  if (ival==1) {
	    pizza_time=time;
	    pizza_status=READY;
	  }

	  break;
#endif

	case READY:

	  break;

	case P1_POS_VIOLATION:
	case P2_POS_VIOLATION:
	case P3_POS_VIOLATION:
	case P4_POS_VIOLATION:
	case P1_VEL_VIOLATION:
	case P2_VEL_VIOLATION:
	case P3_VEL_VIOLATION:
	case P4_VEL_VIOLATION:

	  break;

/*	case P1_CLIPPING:*/
/*	case P2_CLIPPING:*/
/*	case P3_CLIPPING:*/
/*	case P4_CLIPPING:*/

/*	  break; */

	case STOP:

	  break;

	}
}
 




#define MDL_DERIVATIVES
static void mdlDerivatives(SimStruct *S)
{
	real_T *dx=ssGetdX(S);
	real_T *x=ssGetContStates(S);
	InputRealPtrsType uPtrs=ssGetInputPortRealSignalPtrs(S,0);

	int_T *piwrk=ssGetIWork(S);
	real_T *prwrk=ssGetRWork(S);

	int_T i,istat;
	real_T time;

/*	current time */

	time=ssGetT(S);

	istat=(int_T) pizza_status;
	switch (istat)
	{
	case WAITING_FOR_START:

	  for (i=0;i<NSTATES;i++) {
	    dx[i]=0.0;
	  }

	  break;

	case HOMING:

	  pizza_homing_dif(dx,x,&pizza_pos(0),time,S);

	  break;

        case MOVING:

	  pizza_moving_dif(dx,x,&pizza_pos(0),time,S);

	  break;

	case READY:

	  pizza_ready_dif(dx,x,time);

	  break;

	case P1_POS_VIOLATION:
	case P2_POS_VIOLATION:
	case P3_POS_VIOLATION:
	case P4_POS_VIOLATION:
	case P1_VEL_VIOLATION:
	case P2_VEL_VIOLATION:
	case P3_VEL_VIOLATION:
	case P4_VEL_VIOLATION:

	  pizza_violation_dif(dx,x,&pizza_pos(0),time);

	  break;

/*	case P1_CLIPPING: */
/*	case P2_CLIPPING: */
/*	case P3_CLIPPING: */
/*	case P4_CLIPPING: */

/*	  break; */

	case STOP:

	  break;

	}

#ifdef PIZZAMODEL

        for (i=0;i<4;i++) {
	  dx[8+i]=x[12+i];
	  dx[12+i]=-fric(i)*(2.0/PI)*atan(1000.0*x[12+i])-visc(i)*x[12+i]+gain(i)*us(i);
	}
/*	add some gravity... */
	dx[12]=dx[12]+1/0.48;


#endif
}





static void mdlTerminate(SimStruct *S)
{
#ifdef IS_WIN32
#ifdef PIZZAVIEW
	StartupProgram("fclient.exe");
#endif
#endif
}

#ifdef  MATLAB_MEX_FILE    /* Is this file being compiled as a MEX-file? */
#include "simulink.c"      /* MEX-file interface mechanism */
#else
#include "cg_sfun.h"       /* Code generation registration function */
#endif
