
/*#define mexPrintf printf*/

/* jogging and point-to-point motion based on third-degree setpoint profile */

/* Rene' van de Molengraft, April, 14th, 2000 */

/* saves settings for four different axes */

/* time t is the absolute time */





/* shortcuts for jogging variables in workspace */

#define t0(element)		prwrk[piwrk[JOG_IDX]+element]
#define t1(element)		prwrk[piwrk[JOG_IDX+1]+element]
#define t2(element)		prwrk[piwrk[JOG_IDX+2]+element]
#define t3(element)		prwrk[piwrk[JOG_IDX+3]+element]
#define t4(element)		prwrk[piwrk[JOG_IDX+4]+element]
#define t5(element)		prwrk[piwrk[JOG_IDX+5]+element]
#define t6(element)		prwrk[piwrk[JOG_IDX+6]+element]
#define t7(element)		prwrk[piwrk[JOG_IDX+7]+element]
#define s0(element)		prwrk[piwrk[JOG_IDX+8]+element]
#define s1(element)		prwrk[piwrk[JOG_IDX+9]+element]
#define s2(element)		prwrk[piwrk[JOG_IDX+10]+element]
#define s3(element)		prwrk[piwrk[JOG_IDX+11]+element]
#define s4(element)		prwrk[piwrk[JOG_IDX+12]+element]
#define s5(element)		prwrk[piwrk[JOG_IDX+13]+element]
#define s6(element)		prwrk[piwrk[JOG_IDX+14]+element]
#define s7(element)		prwrk[piwrk[JOG_IDX+15]+element]
#define v0(element)		prwrk[piwrk[JOG_IDX+16]+element]
#define v1(element)		prwrk[piwrk[JOG_IDX+17]+element]
#define v2(element)		prwrk[piwrk[JOG_IDX+18]+element]
#define v3(element)		prwrk[piwrk[JOG_IDX+19]+element]
#define v4(element)		prwrk[piwrk[JOG_IDX+20]+element]
#define v5(element)		prwrk[piwrk[JOG_IDX+21]+element]
#define v6(element)		prwrk[piwrk[JOG_IDX+22]+element]
#define v7(element)		prwrk[piwrk[JOG_IDX+23]+element]
#define a0(element)		prwrk[piwrk[JOG_IDX+24]+element]
#define a1(element)		prwrk[piwrk[JOG_IDX+25]+element]
#define a2(element)		prwrk[piwrk[JOG_IDX+26]+element]
#define a3(element)		prwrk[piwrk[JOG_IDX+27]+element]
#define a4(element)		prwrk[piwrk[JOG_IDX+28]+element]
#define a5(element)		prwrk[piwrk[JOG_IDX+29]+element]
#define a6(element)		prwrk[piwrk[JOG_IDX+30]+element]
#define a7(element)		prwrk[piwrk[JOG_IDX+31]+element]
#define delta(element)		prwrk[piwrk[JOG_IDX+32]+element]
#define gamma(element)		prwrk[piwrk[JOG_IDX+33]+element]
#define jerk(element)		prwrk[piwrk[JOG_IDX+34]+element]
#define idir(element)		prwrk[piwrk[JOG_IDX+35]+element]
#define xstrt(element)		prwrk[piwrk[JOG_IDX+36]+element]





void jog_ini(real_T xstart, real_T tstart, real_T vdes, real_T tdes, real_T maxjerk, int_T i, SimStruct *S)
{

/*
	input arguments

	xstart	: start position
	tstart	: start time
	vdes	: desired jogging speed
	tdes	: time to reach vdes
	maxjerk	: jerk in acceleration phase
	i	: axis id
*/

	real_T det;

	int_T *piwrk=ssGetIWork(S);
	real_T *prwrk=ssGetRWork(S);

	if (vdes>=0.0) {
	  idir(i)=1.0;
	} else {
	  idir(i)=-1.0;
	  vdes=-vdes;
	}

	if (maxjerk>=0.0) {
	  jerk(i)=maxjerk;
	} else {
	  jerk(i)=-maxjerk;
	}

/*	compute jerk period delta */

	det=tdes*tdes*jerk(i)*jerk(i)-4.0*jerk(i)*vdes;

	if (det<0) {
	  ssSetErrorStatus(S, "JOG_INI: vdes cannot be reached (increase jerk and/or tdes).");
	} else {
          delta(i)=(tdes*jerk(i)-sqrt(det))/(2.0*jerk(i));
   
/*	  compute acceleration period gamma */

          gamma(i)=tdes-2.0*delta(i);
   
/*	  compute switching times */
   
	  t0(i)=tstart;
	  t1(i)=t0(i)+delta(i);
	  t2(i)=t1(i)+gamma(i);
	  t3(i)=t2(i)+delta(i);

/*	  t4, t5, t6 and t7 equal infinity at startup */

	  t4(i)=100000.0;
	  t5(i)=100000.0;
	  t6(i)=100000.0;
	  t7(i)=100000.0;
   
/*	  compute reference values at switching times */
   
	  xstrt(i)=xstart;
	  s0(i)=0.0;
   
	  a1(i)=jerk(i)*(t1(i)-t0(i));
	  v1(i)=0.5*jerk(i)*(t1(i)-t0(i))*(t1(i)-t0(i));
	  s1(i)=s0(i)+jerk(i)*(t1(i)-t0(i))*(t1(i)-t0(i))*(t1(i)-t0(i))/6.0;
   
  	  a2(i)=a1(i);
	  v2(i)=v1(i)+a1(i)*(t2(i)-t1(i));
	  s2(i)=s1(i)+v1(i)*(t2(i)-t1(i))+0.5*a1(i)*(t2(i)-t1(i))*(t2(i)-t1(i));
   
	  a3(i)=a2(i)-jerk(i)*(t3(i)-t2(i));
	  v3(i)=v2(i)+a2(i)*(t3(i)-t2(i))-0.5*jerk(i)*(t3(i)-t2(i))*(t3(i)-t2(i));
	  s3(i)=s2(i)+v2(i)*(t3(i)-t2(i))+0.5*a2(i)*(t3(i)-t2(i))*(t3(i)-t2(i))
                -jerk(i)*(t3(i)-t2(i))*(t3(i)-t2(i))*(t3(i)-t2(i))/6.0;
	}
}





void jog_get(real_T *qref, real_T *vref, real_T *aref, real_T t, int_T i, SimStruct *S)
{

/*	i : axis id */

	int_T *piwrk=ssGetIWork(S);
	real_T *prwrk=ssGetRWork(S);

	if (t<=t0(i)) {
	  aref[0]=0;
	  vref[0]=0;
	  qref[0]=s0(i);
        } else if (t<=t1(i)) {
	  aref[0]=jerk(i)*(t-t0(i));
	  vref[0]=0.5*jerk(i)*(t-t0(i))*(t-t0(i));
          qref[0]=s0(i)+jerk(i)*(t-t0(i))*(t-t0(i))*(t-t0(i))/6.0;
        } else if (t<=t2(i)) {
	  aref[0]=a1(i);
	  vref[0]=v1(i)+a1(i)*(t-t1(i));
	  qref[0]=s1(i)+v1(i)*(t-t1(i))+0.5*a1(i)*(t-t1(i))*(t-t1(i));
        } else if (t<=t3(i)) {
	  aref[0]=a2(i)-jerk(i)*(t-t2(i));
	  vref[0]=v2(i)+a2(i)*(t-t2(i))-0.5*jerk(i)*(t-t2(i))*(t-t2(i));
          qref[0]=s2(i)+v2(i)*(t-t2(i))+0.5*a2(i)*(t-t2(i))*(t-t2(i))
                  -jerk(i)*(t-t2(i))*(t-t2(i))*(t-t2(i))/6.0;
        } else if (t<=t4(i)) {
	  aref[0]=0;
	  vref[0]=v3(i);
	  qref[0]=s3(i)+v3(i)*(t-t3(i));
        } else if (t<=t5(i)) {
	  aref[0]=-jerk(i)*(t-t4(i));
	  vref[0]=v4(i)-0.5*jerk(i)*(t-t4(i))*(t-t4(i));
	  qref[0]=s4(i)+v4(i)*(t-t4(i))-jerk(i)*(t-t4(i))*(t-t4(i))*(t-t4(i))/6.0;
        } else if (t<=t6(i)) {
	  aref[0]=a5(i);
	  vref[0]=v5(i)+a5(i)*(t-t5(i));
	  qref[0]=s5(i)+v5(i)*(t-t5(i))+0.5*a5(i)*(t-t5(i))*(t-t5(i));
        } else if (t<=t7(i)) {
	  aref[0]=a6(i)+jerk(i)*(t-t6(i));
	  vref[0]=v6(i)+a6(i)*(t-t6(i))+0.5*jerk(i)*(t-t6(i))*(t-t6(i));
	  qref[0]=s6(i)+v6(i)*(t-t6(i))+0.5*a6(i)*(t-t6(i))*(t-t6(i))
		  +jerk(i)*(t-t6(i))*(t-t6(i))*(t-t6(i))/6.0;
	} else {
	  aref[0]=0.0;
	  vref[0]=0.0;
	  qref[0]=s7(i);
	}

	if (idir(i)==-1.0) {
	  aref[0]=-aref[0];
	  vref[0]=-vref[0];
	  qref[0]=-qref[0];
	}
	qref[0]=qref[0]+xstrt(i);
}





void jog_stop(real_T t, int_T i, SimStruct *S)
{

/*	i : axis id */

	int_T *piwrk=ssGetIWork(S);
	real_T *prwrk=ssGetRWork(S);

	t4(i)=t;
	t5(i)=t4(i)+delta(i);
	t6(i)=t5(i)+gamma(i);
	t7(i)=t6(i)+delta(i);

	a4(i)=0.0;
	v4(i)=v3(i);
	s4(i)=s3(i)+v3(i)*(t4(i)-t3(i));

	a5(i)=-jerk(i)*(t5(i)-t4(i));
	v5(i)=v4(i)-0.5*jerk(i)*(t5(i)-t4(i))*(t5(i)-t4(i));
	s5(i)=s4(i)+v4(i)*(t5(i)-t4(i))-jerk(i)*(t5(i)-t4(i))*(t5(i)-t4(i))*(t5(i)-t4(i))/6.0;
   
	a6(i)=a5(i);
	v6(i)=v5(i)+a5(i)*(t6(i)-t5(i));
	s6(i)=s5(i)+v5(i)*(t6(i)-t5(i))+0.5*a5(i)*(t6(i)-t5(i))*(t6(i)-t5(i));
   
	a7(i)=a6(i)+jerk(i)*(t7(i)-t6(i));
	v7(i)=v6(i)+a6(i)*(t7(i)-t6(i))+0.5*jerk(i)*(t7(i)-t6(i))*(t7(i)-t6(i));
	s7(i)=s6(i)+v6(i)*(t7(i)-t6(i))+0.5*a6(i)*(t7(i)-t6(i))*(t7(i)-t6(i))
	      +jerk(i)*(t7(i)-t6(i))*(t7(i)-t6(i))*(t7(i)-t6(i))/6.0;
}





int_T jog_status(real_T t, int_T i, SimStruct *S)
{

/*	i : axis id */

	int_T *piwrk=ssGetIWork(S);
	real_T *prwrk=ssGetRWork(S);

	if (t>t7(i)+1.0) {
	  return 1;
	} else {
	  return 0;
	}
}





void p2p_ini(real_T xstart, real_T tstart, real_T xend, real_T vdes, real_T tdes, real_T maxjerk, int_T i, SimStruct *S)
{

/*
	input arguments

	xstart	: start position
	tstart	: start time
	xend	: end position
	vdes	: desired jogging speed
	tdes	: time to reach vdes
	maxjerk	: jerk in acceleration phase
	i	: axis id
*/

	real_T det,disp;

	int_T *piwrk=ssGetIWork(S);
	real_T *prwrk=ssGetRWork(S);

#ifdef PIZZADEBUG
	mexPrintf("p2p_ini: xstart = %f\n",xstart);
	mexPrintf("p2p_ini: tstart = %f\n",tstart);
	mexPrintf("p2p_ini: xend = %f\n",xend);
	mexPrintf("p2p_ini: vdes = %f\n",vdes);
	mexPrintf("p2p_ini: tdes = %f\n",tdes);
	mexPrintf("p2p_ini: maxjerk = %f\n",maxjerk);
	mexPrintf("p2p_ini: iaxis = %d\n",i);
#endif

	disp=xend-xstart;
	if (disp<0.0) {
	  disp=-disp;
	}

	if (vdes>=0.0) {
	  idir(i)=1.0;
	} else {
	  idir(i)=-1.0;
	  vdes=-vdes;
	}

	if (maxjerk>=0.0) {
	  jerk(i)=maxjerk;
	} else {
	  jerk(i)=-maxjerk;
	}

/*	compute jerk period delta */

	det=tdes*tdes*jerk(i)*jerk(i)-4.0*jerk(i)*vdes;

	if (det<0) {
	  ssSetErrorStatus(S, "JOG_INI: vdes cannot be reached (increase jerk and/or tdes).");
	} else {
          delta(i)=(tdes*jerk(i)-sqrt(det))/(2.0*jerk(i));
   
/*	  compute acceleration period gamma */

          gamma(i)=tdes-2.0*delta(i);
   
/*	  compute switching times */
   
	  t0(i)=tstart;
	  t1(i)=t0(i)+delta(i);
	  t2(i)=t1(i)+gamma(i);
	  t3(i)=t2(i)+delta(i);

/*	  compute t4 */

	  t4(i)=t3(i)+(disp-2.0*jerk(i)*delta(i)*delta(i)*delta(i)
		-3.0*jerk(i)*delta(i)*delta(i)*gamma(i)-jerk(i)*delta(i)*gamma(i)*gamma(i))/vdes;
	  if (t4(i)<t3(i)) {
	    ssSetErrorStatus(S, "P2P_INI: vdes too high for displacement (decrease vdes).");
	  } else {
	    t5(i)=t4(i)+delta(i);
	    t6(i)=t5(i)+gamma(i);
	    t7(i)=t6(i)+delta(i);   
   
/*	    compute reference values at switching times */
   
	    xstrt(i)=xstart;
	    s0(i)=0.0;
   
	    a1(i)=jerk(i)*(t1(i)-t0(i));
	    v1(i)=0.5*jerk(i)*(t1(i)-t0(i))*(t1(i)-t0(i));
	    s1(i)=s0(i)+jerk(i)*(t1(i)-t0(i))*(t1(i)-t0(i))*(t1(i)-t0(i))/6.0;
   
  	    a2(i)=a1(i);
	    v2(i)=v1(i)+a1(i)*(t2(i)-t1(i));
	    s2(i)=s1(i)+v1(i)*(t2(i)-t1(i))+0.5*a1(i)*(t2(i)-t1(i))*(t2(i)-t1(i));
   
	    a3(i)=a2(i)-jerk(i)*(t3(i)-t2(i));
	    v3(i)=v2(i)+a2(i)*(t3(i)-t2(i))-0.5*jerk(i)*(t3(i)-t2(i))*(t3(i)-t2(i));
	    s3(i)=s2(i)+v2(i)*(t3(i)-t2(i))+0.5*a2(i)*(t3(i)-t2(i))*(t3(i)-t2(i))
		  -jerk(i)*(t3(i)-t2(i))*(t3(i)-t2(i))*(t3(i)-t2(i))/6.0;

	    a4(i)=0.0;
	    v4(i)=v3(i);
	    s4(i)=s3(i)+v3(i)*(t4(i)-t3(i));

	    a5(i)=-jerk(i)*(t5(i)-t4(i));
	    v5(i)=v4(i)-0.5*jerk(i)*(t5(i)-t4(i))*(t5(i)-t4(i));
	    s5(i)=s4(i)+v4(i)*(t5(i)-t4(i))-jerk(i)*(t5(i)-t4(i))*(t5(i)-t4(i))*(t5(i)-t4(i))/6.0;
   
	    a6(i)=a5(i);
	    v6(i)=v5(i)+a5(i)*(t6(i)-t5(i));
	    s6(i)=s5(i)+v5(i)*(t6(i)-t5(i))+0.5*a5(i)*(t6(i)-t5(i))*(t6(i)-t5(i));
   
	    a7(i)=a6(i)+jerk(i)*(t7(i)-t6(i));
	    v7(i)=v6(i)+a6(i)*(t7(i)-t6(i))+0.5*jerk(i)*(t7(i)-t6(i))*(t7(i)-t6(i));
	    s7(i)=s6(i)+v6(i)*(t7(i)-t6(i))+0.5*a6(i)*(t7(i)-t6(i))*(t7(i)-t6(i))
		  +jerk(i)*(t7(i)-t6(i))*(t7(i)-t6(i))*(t7(i)-t6(i))/6.0;
	  }
	}
}





void p2p_get(real_T *qref, real_T *vref, real_T *aref, real_T t, int_T i, SimStruct *S)
{
	jog_get(qref,vref,aref,t,i,S);
}





int_T p2p_status(real_T t, int_T i, SimStruct *S)
{
	return jog_status(t,i,S);
}
